#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <unistd.h>
#include <ctype.h>
#include <getopt.h>

char buf[8192];
struct jacosub {
	int timer;
	char *text;
} *jacos;
int njacos,pjacos;

char *trim(char *str)
{
	char *d,*e;
	while (*str && isspace(*str)) str++;
	for (d=e=str;*d;d++) if (!isspace(*d)) e=d+1;
	*e=0;
	return str;
}

char *volfile;

void read_jacos(char *fname)
{
	FILE *f;
	int hd;
	char *c,*c1,*c2;
	f=fopen(fname,"r");
	if (!f) {
		perror(fname);
		exit(0);
	}
	while(fgets(buf,8192,f)) {
		if (!isdigit(buf[0])) continue;
		hd=36000*strtol(buf,&c,10);
		if (*c++!=':') continue;
		hd+=600*strtol(c,&c,10);
		if (*c++!=':') continue;
		hd+=10*strtol(c,&c,10);
		if (*c++!='.') continue;
		hd+=strtol(c,&c,10)/10;
		while (*c && isspace(*c)) c++;
		while (*c && !isspace(*c)) c++;
		while (c1=strchr(c,'{')) {
			c2=strchr(c1,'}');
			if (!c2) {
				*c1=0;
				break;
			}
			strcpy(c1,c2+1);
		}
		c=trim(c);
		if (!*c) continue;
		if (!pjacos) {
			pjacos=1024;
			jacos=malloc(sizeof(*jacos)*pjacos);
		}
		else if (njacos >= pjacos) {
			pjacos+=1024;
			jacos=realloc(jacos,sizeof(*jacos)*pjacos);
		}
		for (c1=c2=c;*c1;) {
			if (*c1 != '\\') {
				*c2++=*c1++;
				continue;
			}
			c1++;
			*c2++=' ';
			if (*c1) c1++;
		}
		*c2=0;
		jacos[njacos].timer=hd;
		jacos[njacos++].text=strdup(c);
	}
	fclose(f);
}

static int blank=1;

int get_pos(void)
{
	fd_set rdfs,exfs;
	char znak;
	struct timeval tv;
	int retval;
	static int seconds,tenths;
	
	FD_ZERO(&rdfs);
	FD_SET(0,&rdfs);
	FD_ZERO(&exfs);
	FD_SET(0,&exfs);
	tv.tv_sec=0;
	tv.tv_usec=200000;
	retval=select(1,&rdfs,NULL,&exfs,&tv);
	if (retval == -1) {
		perror("select");
		exit(1);
	}
	if (!retval) return -1;
	if (FD_ISSET(0,&exfs)) {
		fprintf(stderr,"Good bye\n");
		exit(0);
	}
	if (!FD_ISSET(0,&rdfs)) return -1;
	if (read(0,&znak,1)!=1) {
		fprintf(stderr,"Good bye\n");
		exit(0);
	}
	if (blank==4) {
		if (isdigit(znak)) {
			seconds=(10*seconds)+(znak & 15);
			return -1;
		}
		if (znak=='.') {
			blank=5;
			return -1;
		}
		if (isspace(znak)) blank=1;
		else blank=0;
		return seconds*10+tenths;
	}
	if (blank==5) {
		if (isdigit(znak)) {
			tenths=znak & 15;
			blank=0;
			return seconds*10+tenths;
		}
		if (isspace(znak)) blank=1;
		else blank=0;
		return seconds*10+tenths;
	}
	if (blank==1) {
		if (znak != 'V') blank=0;
		else blank=2;
		return -1;
	}
	if (blank==2) {
		if (znak != ':') blank=0;
		else blank=3;
		return -1;
	}
	if (blank==3) {
		if (isspace(znak)) return -1;
		if (!isdigit(znak)) blank=0;
		else {
			seconds=znak & 15;
			tenths=0;
			blank=4;
		}
		return -1;
	}
	if (isspace(znak)) blank=1;
	else blank=0;
	return -1;	
			
			
}

pid_t say_pid=0;
int last_pos=-1;
int last_line=-1;

int qlen;
char *queue[10];

void stop_speech(void)
{
	qlen=0;
	if (say_pid) kill(-say_pid,SIGTERM);
}

void say_direct(char *str)
{
	FILE *f,*vf;
	char cmdline[256];
	say_pid=fork();
	if (say_pid <0) {
		perror("fork");
		exit(1);
	}
	if (say_pid) return;
	setpgid(0,0);
	close(1);
	close(2);
	strcpy(cmdline,"milena_say -T -i -E UTF-8");
	vf=fopen(volfile,"r");
	if (vf) {
		char buf[32];
		if (fgets(buf,32,vf)) {
			sprintf(cmdline+strlen(cmdline)," -v %f",strtod(buf,NULL));
		}
		fclose(vf);
	}
	f=popen(cmdline,"w");
	fprintf(f,"%s\n",str);
	pclose(f);
	exit(0);
}

void say_queue(void)
{
	int i;
	if (qlen<=0) return;
	say_direct(queue[0]);
	for (i=1;i<qlen;i++) queue[i-1]=queue[i];
	qlen--;
}

void say(char *str)
{
	if (!say_pid) {
		say_direct(str);
		return;
	}
	if (qlen>=10) return;
	queue[qlen++]=str;
	
}

int tod(void)
{
	struct timeval tv;
	gettimeofday(&tv,NULL);
	return (tv.tv_sec-1200000000)*10+tv.tv_usec/100000;
}

void help(char *name)
{
	fprintf(stderr,"Uzycie: %s [-v plik_glosnosci] napisy.jss\n",name);
	exit(0);
}

main(int argc,char *argv[])
{
	int pos;
	int timmy,tammy;
	int started=0;

	for(;;) {
		int c=getopt(argc,argv,"hv:");
		if (c<0) break;
		if (c=='v') {
			volfile=optarg;
			continue;
		}
		help(argv[0]);
	}

	if (optind != argc-1) {
		help(argv[0]);
		exit(1);
	}
	read_jacos(argv[optind]);
	for (;;) {
		if (say_pid) {
			pid_t t=waitpid(say_pid,NULL,WNOHANG);
			if (t==say_pid) {
				say_pid=0;
				say_queue();
			}
		}
		pos=get_pos();
		if (pos>=0) {
			int toll=0;
			int i;
			if (!started) {
				started=1;
			}
			if (pos > last_pos+10) {
				stop_speech();
				toll=-3;
			}
			else if (pos < last_pos) {
				stop_speech();
				last_line=-1;
			}
			last_pos=pos;
			for (i=last_line+1;i<njacos;i++) {
				if (jacos[i].timer > last_pos) break;
				if (jacos[i].timer >= last_pos + toll) {
					last_line=i;
					say(jacos[i].text);
					break;
				}
			}
			timmy=tod();

		}
		else if (started) {
			tammy=tod();
			if (tammy> timmy+2) {
				stop_speech();
				started=0;
			}
		}
	}
}
